from icommandlib import messages as message
from icommandlib.handle import IProcessHandle
from icommandlib.utils import stripshot
from icommandlib import exceptions
import threading
import psutil
import queue


class IProcess(object):
    """
    An interactive process.

        Start process by initializing with an commandlib.Command object.
    """
    def __init__(self, icommand):
        self._icommand = icommand
        self._order_queue = queue.Queue()
        self._event_queue = queue.Queue()
        self._handle = threading.Thread(
            target=IProcessHandle,
            args=(icommand, self._order_queue, self._event_queue)
        )
        self._handle.start()

        self._running_process = self._wait_for_event(message.ProcessStartedMessage)

        self._pid = self._running_process._pid
        self._master_fd = self._running_process._stdin
        self._async_send = self._wait_for_event(message.AsyncSendMethodMessage)

        self._final_screenshot = None
        self._running = True
        self._exit_code = None

    def _check_events(self, expected=None):
        try:
            response = self._event_queue.get(block=False)
        except queue.Empty:
            response = None

        if response is not None:
            return self._handle_event(response, of_kind=expected)

    def _handle_event(self, response, of_kind=None):
        if isinstance(response, message.ExceptionMessage):
            raise response.value
        if isinstance(response, message.TimeoutMessage):
            raise exceptions.IProcessTimeout(
                "Timed out after {0} seconds:\n\n{1}".format(
                    response.after,
                    stripshot(response.screenshot),
                )
            )
        if isinstance(response, message.ExitMessage):
            self._running = False
            self._pid = None
            self._exit_code = response.value.exit_code
            self._final_screenshot = response.value.screenshot
            if of_kind != message.ExitMessage:
                raise exceptions.UnexpectedExit(
                    response.value.exit_code,
                    response.value.screenshot,
                )
        if not isinstance(response, of_kind):
            raise Exception(
                "Threading error expected {0} got {1}".format(
                    type(of_kind), type(of_kind)
                )
            )
        return response.value

    def _wait_for_event(self, of_kind, timeout=None):
        try:
            response = self._event_queue.get(timeout=timeout)
        except queue.Empty:
            raise exceptions.IProcessTimeout(
                "Timed out waiting for exit."
            )
        return self._handle_event(response, of_kind=of_kind)

    @property
    def pid(self):
        self._check_events()
        return self._pid

    @property
    def running(self):
        self._check_events(expected=message.ExitMessage)
        return self._running

    @property
    def exit_code(self):
        self._check_events(expected=message.ExitMessage)
        return self._exit_code

    @property
    def psutil(self):
        """
        Return psutil.Process object from current process,
        if psutil is installed.
        """
        return psutil.Process(self._pid)

    def wait_until(self, condition_function, timeout=None):
        if self._running:
            self._check_events()
            self._order_queue.put(message.Condition(
                condition_function, timeout
            ))
            self._async_send()
            self._wait_for_event(message.OutputMatched)
        else:
            raise exceptions.AlreadyExited(
                self._exit_code,
                self._final_screenshot,
            )

    def wait_until_output_contains(self, text, timeout=None):
        """
        Wait until the totality of the output of the process contains
        at least one instance of 'text'.
        """
        self.wait_until(
            lambda iscreen: text in iscreen.raw_bytes.decode('utf8'),
            timeout,
        )

    def wait_until_on_screen(self, text, timeout=None):
        """
        Waits until the text specified appears on the process's terminal
        screen.
        """
        def on_screen(iscreen):
            return len([
                line for line in iscreen.display if text in line
            ]) > 0

        self.wait_until(on_screen, timeout)

    def send_keys(self, text):
        """
        Send keys to the terminal process.
        """
        if self._running:
            self._check_events()
            self._order_queue.put(message.KeyPresses(text.encode('utf8')))
            self._async_send()
        else:
            raise exceptions.AlreadyExited(
                self._exit_code,
                self._final_screenshot,
            )

    def screenshot(self):
        """
        Get a 'screenshot' of the terminal window of a running or finished
        process.

        Use stripshot() to clean the whitespace from the right and bottom.
        """
        if self._final_screenshot is not None:
            return self._final_screenshot
        else:
            self._order_queue.put(message.TakeScreenshot())
            self._async_send()
            return self._wait_for_event(message.Screenshot)

    def stripshot(self):
        """
        Get a stripped screenshot of the terminal window of the running process.

        i.e. all of the whitepace to the right and bottom will be cut.
        """
        return stripshot(self.screenshot())

    def wait_for_finish(self, timeout=None):
        """
        Wait until the process has finished.
        """
        self._wait_for_event(message.ExitMessage, timeout=timeout)

    def wait_for_successful_exit(self, timeout=None):
        """
        Wait until the process exits successfully. Raises exception
        if the process is still open after timeout or it exits
        with an exit code other than 0.
        """
        response = self._wait_for_event(message.ExitMessage, timeout=timeout)

        if response.exit_code != 0:
            raise exceptions.ExitWithError(
                response.exit_code,
                stripshot(response.screenshot),
            )

        return response

    def kill(self):
        """
        Send a kill -9 to the process and all of its descendants.
        """
        if self._running:
            self._check_events()
            self._order_queue.put(message.KillProcess())
            self._async_send()
            self._wait_for_event(message.ProcessKilled)
        else:
            raise exceptions.AlreadyExited(
                self._exit_code,
                self._final_screenshot,
            )



from hitchstory import StoryCollection, GivenDefinition, GivenProperty, validate
from hitchstory import InfoDefinition, InfoProperty
from hitchstory import BaseEngine, no_stacktrace_for, HitchStoryException
from hitchrun import expected, DIR
from commandlib import Command, CommandError
from pathquery import pathquery
from strictyaml import MapPattern, Map, Str, Float, Seq, Bool
from hitchrunpy import HitchRunPyException, ExamplePythonCode, ExpectedExceptionMessageWasDifferent
from commandlib import python
import hitchpylibrarytoolkit
from templex import Templex
import signal


class Engine(BaseEngine):
    given_definition = GivenDefinition(
        files=GivenProperty(MapPattern(Str(), Str())),
        variables=GivenProperty(MapPattern(Str(), Str())),
        python_version=GivenProperty(Str()),
        setup=GivenProperty(Str()),
        code=GivenProperty(Str()),
    )

    info_definition = InfoDefinition(
        known_failure=InfoProperty(Bool()),
    )

    def __init__(self, pathgroup, settings):
        self.path = pathgroup
        self.settings = settings

    def set_up(self):
        self.path.state = self.path.gen.joinpath("state")
        if self.path.state.exists():
            self.path.state.rmtree(ignore_errors=True)
        self.path.state.mkdir()

        for filename, text in self.given.get("files", {}).items():
            filepath = self.path.state.joinpath(filename)
            if not filepath.dirname().exists():
                filepath.dirname().mkdir()
            filepath.write_text(str(text))
            filepath.chmod("u+x")

        for filename, text in self.given.get("variables", {}).items():
            filepath = self.path.state.joinpath(filename)
            if not filepath.dirname().exists():
                filepath.dirname().mkdir()
            filepath.write_text(str(text))

        self.path.key.joinpath("code_that_does_things.py").copy(self.path.state)

        self.python = hitchpylibrarytoolkit.project_build(
            "icommandlib",
            self.path,
            self.given["python version"],
        ).bin.python

        self.example_py_code = ExamplePythonCode(
            self.python, self.path.state,
        ).with_setup_code(
            self.given.get('setup', '')
        ).with_code(
            self.given.get('code', '')
        ).with_timeout(4.0)

    @no_stacktrace_for(HitchRunPyException)
    def run_code(self):
        self.result = self.example_py_code.run()

    @no_stacktrace_for(HitchRunPyException)
    def start_code(self):
        self.running_python = self.example_py_code.running_code()

    def pause_for_half_a_second(self):
        import time
        time.sleep(0.5)

    def send_signal_and_wait_for_finish(self, signal_name):
        SIGNAL_NAMES_TO_NUMBERS = {
            name: getattr(signal, name) for name in dir(signal)
            if name.startswith('SIG') and '_' not in name
        }
        self.running_python.iprocess.psutil._send_signal(
            int(SIGNAL_NAMES_TO_NUMBERS[signal_name])
        )
        self.running_python.iprocess.wait_for_finish()

    @no_stacktrace_for(HitchRunPyException)
    @validate(
        exception_type=Map({"in python 2": Str(), "in python 3": Str()}) | Str(),
        message=Map({"in python 2": Str(), "in python 3": Str()}) | Str(),
    )
    def raises_exception(self, exception_type=None, message=None):
        """
        Expect an exception.
        """
        differential = False

        if exception_type is not None:
            if not isinstance(exception_type, str):
                differential = True
                exception_type = exception_type['in python 2']\
                    if self.given['python version'].startswith("2")\
                    else exception_type['in python 3']

        if message is not None:
            if not isinstance(message, str):
                differential = True
                message = message['in python 2']\
                    if self.given['python version'].startswith("2")\
                    else message['in python 3']

        try:
            result = self.example_py_code.expect_exceptions().run()
            result.exception_was_raised(exception_type, message)
        except ExpectedExceptionMessageWasDifferent as error:
            if self.settings.get("overwrite artefacts") and not differential:
                self.current_step.update(message=error.actual_message)
            else:
                raise

    @validate(from_filenames=Seq(Str()))
    def processes_not_alive(self, from_filenames=None):
        still_alive = []
        for from_filename in from_filenames:
            import psutil
            pid = int(self.path.state.joinpath(from_filename).bytes().decode('utf8').strip())
            try:
                proc = psutil.Process(pid)
                proc.kill()
                still_alive.append(from_filename)
            except psutil.NoSuchProcess:
                pass
        if len(still_alive) > 0:
            raise Exception("Processes from {0} still alive.".format(', '.join(still_alive)))

    def touch_file(self, filename):
        self.path.state.joinpath(filename).write_text("\nfile touched!", append=True)

    def _will_be(self, content, text, reference, changeable=None):
        if text is not None:
            if content.strip() == text.strip():
                return
            else:
                raise RuntimeError("Expected to find:\n{0}\n\nActual output:\n{1}".format(
                    text,
                    content,
                ))

        artefact = self.path.key.joinpath(
            "artefacts", "{0}.txt".format(reference.replace(" ", "-").lower())
        )
        text = artefact.text()

        if not artefact.exists():
            artefact.write_text(content)
        else:
            if self.settings.get('overwrite artefacts'):
                if artefact.bytes().decode('utf8') != content:
                    artefact.write_text(content)
            else:
                Templex(text).assert_match(content)

    def file_contents_will_be(self, filename, text=None, reference=None, changeable=None):
        output_contents = self.path.state.joinpath(filename).bytes().decode('utf8')
        self._will_be(output_contents, text, reference, changeable)

    def output_will_be(self, text=None, reference=None, changeable=None):
        output_contents = self.path.state.joinpath("output.txt").bytes().decode('utf8')
        self._will_be(output_contents, text, reference, changeable)

    @validate(seconds=Float())
    def sleep(self, seconds):
        import time
        time.sleep(float(seconds))

    def on_success(self):
        if self.settings.get("overwrite artefacts"):
            self.new_story.save()


@expected(HitchStoryException)
def bdd(*words):
    """
    Run story in BDD mode that matches keywords (e.g. tdd wait finished)
    """
    StoryCollection(
        pathquery(DIR.key).ext("story"), Engine(DIR, {"overwrite artefacts": True})
    ).shortcut(*words).play()


def regression():
    """
    Run all stories.
    """
    StoryCollection(
        pathquery(DIR.key).ext("story"), Engine(DIR, {})
    ).only_uninherited().filter(
        lambda story: not story.info.get('known_failure')
    ).ordered_by_name().play()


def rewrite():
    """
    Run all stories and rewrite any with different output.
    """
    StoryCollection(
        pathquery(DIR.key).ext("story"), Engine(DIR, {"overwrite artefacts": True})
    ).only_uninherited().ordered_by_name().play()


def deploy(version):
    """
    Deploy to pypi as specified version.
    """
    NAME = "icommandlib"
    git = Command("git").in_dir(DIR.project)
    version_file = DIR.project.joinpath("VERSION")
    old_version = version_file.bytes().decode('utf8')
    if version_file.bytes().decode("utf8") != version:
        DIR.project.joinpath("VERSION").write_text(version)
        git("add", "VERSION").run()
        git("commit", "-m", "RELEASE: Version {0} -> {1}".format(
            old_version,
            version
        )).run()
        git("push").run()
        git("tag", "-a", version, "-m", "Version {0}".format(version)).run()
        git("push", "origin", version).run()
    else:
        git("push").run()

    # Set __version__ variable in __init__.py, build sdist and put it back
    initpy = DIR.project.joinpath(NAME, "__init__.py")
    original_initpy_contents = initpy.bytes().decode('utf8')
    initpy.write_text(
        original_initpy_contents.replace("DEVELOPMENT_VERSION", version)
    )
    python("setup.py", "sdist").in_dir(DIR.project).run()
    initpy.write_text(original_initpy_contents)

    # Upload to pypi
    python(
        "-m", "twine", "upload", "dist/{0}-{1}.tar.gz".format(NAME, version)
    ).in_dir(DIR.project).run()


@expected(CommandError)
def doctest(version="3.5.0"):
    Command(DIR.gen.joinpath("py{0}".format(version), "bin", "python"))(
        "-m", "doctest", "-v", DIR.project.joinpath("icommandlib", "utils.py")
    ).in_dir(DIR.project.joinpath("icommandlib")).run()


def rerun(version="3.5.0"):
    """
    Rerun last example code block with specified version of python.
    """
    Command(DIR.gen.joinpath("py{0}".format(version), "bin", "python"))(
        DIR.gen.joinpath("state", "examplepythoncode.py")
    ).in_dir(DIR.gen.joinpath("state")).run()


def lint():
    """
    Lint all code.
    """
    python("-m", "flake8")(
        DIR.project.joinpath("icommandlib"),
        "--max-line-length=100",
        "--exclude=__init__.py",
    ).run()
    python("-m", "flake8")(
        DIR.key.joinpath("key.py"),
        "--max-line-length=100",
        "--exclude=__init__.py",
    ).run()
    print("Lint success!")



from icommandlib import messages as message
import signal
import psutil
import queue
import pyte
import pyuv
import pty
import os


class IScreen(object):
    """
    Represents the output of a process, both as raw bytes
    and a virtual terminal (screen).
    """
    def __init__(self, screen, raw_byte_output):
        self.screen = screen
        self.raw_bytes = raw_byte_output

    @property
    def display(self):
        return self.screen.display

    @property
    def text(self):
        return u'\n'.join(self.display)


class IProcessHandle(object):
    """
    Starts and then manages with the process directly.

    The IProcessHandle operates in its own thread which operates and
    reports to the main thread in response to the following events:

    * The process exiting of its own accord.
    * Orders from the master thread.
    * A timer (used for timing out waits).
    * Chunks of output spat out at the terminal (fed into virtual terminal).
    """
    def __init__(self, icommand, order_queue, event_queue):
        try:
            self.order_queue = order_queue   # Messages from master thread
            self.event_queue = event_queue   # Messages to master thread

            self.master_fd, self.slave_fd = pty.openpty()

            self.stream = pyte.Stream()
            self.screen = pyte.Screen(icommand.width, icommand.height)
            self.stream.attach(self.screen)
            self.raw_byte_output = b''
            self.timeout = icommand.timeout
            self.task = None

            self.loop = pyuv.Loop.default_loop()

            self.tty_handle = pyuv.TTY(self.loop, self.master_fd, True)
            self.tty_handle.start_read(self.on_tty_read)

            self.sigterm_handle = pyuv.Signal(self.loop)
            self.sigterm_handle.start(self.sigterm_callback, signal.SIGTERM)
            self.sigint_handle = pyuv.Signal(self.loop)
            self.sigint_handle.start(self.sigint_callback, signal.SIGINT)

            self.process_handle = pyuv.Process.spawn(
                self.loop,
                args=icommand._command.arguments,
                env=icommand._command.env,
                cwd=icommand._command.directory,
                exit_callback=self.on_exit,
                stdio=[
                    pyuv.StdIO(
                        fd=self.slave_fd,
                        flags=pyuv.UV_INHERIT_FD,
                    ),
                    pyuv.StdIO(
                        fd=self.slave_fd,
                        flags=pyuv.UV_INHERIT_FD,
                    ),
                    pyuv.StdIO(
                        fd=self.slave_fd,
                        flags=pyuv.UV_INHERIT_FD,
                    ),
                ],
                flags=pyuv.UV_PROCESS_DETACHED,
            )

            self._pid = self.process_handle.pid

            self.event_queue.put(
                message.ProcessStartedMessage(message.RunningProcess(
                    self._pid, self.master_fd
                ))
            )

            self.async_handle = pyuv.Async(self.loop, self.on_thread_callback)
            self.event_queue.put(message.AsyncSendMethodMessage(self.async_handle.send))

            self.timeout_handle = None
            self.reset_timeout()

            self.loop.run()
        except Exception as error:
            self.close_handles()
            self.event_queue.put(message.ExceptionMessage(error))

    def screenshot(self):
        """
        Get a current screenshot of the screen as a string.
        """
        return u"\n".join(line for line in self.screen.display)

    @property
    def psutil(self):
        return psutil.Process(self._pid)

    def on_thread_callback(self, async_handle):
        """
        This is the callback that is triggered when self.async.send()
        (the only threadsafe method on this class) is called.

        It is used to indicate that there's a message waiting on
        self._order_queue to pick up.
        """
        self.check_order_queue()

    def signal_callback(self, handle, signum):
        """
        Callback that is triggered by the parent process receiving
        a sigterm or sigint.
        """
        if signum in (signal.SIGTERM, signal.SIGINT):
            proc = psutil.Process(self.process_handle.pid)
            for descendant in proc.children(recursive=True):
                descendant.kill()
            proc.kill()
        self.check_order_queue()
        self.close_handles()
        self.event_queue.put(
            message.ExitMessage(message.FinishedProcess(
                None,
                9,
                '\n'.join(self.screen.display),
            ))
        )

    def sigterm_callback(self, handle, signum):
        self.signal_callback(handle, signum)

    def sigint_callback(self, handle, signum):
        self.signal_callback(handle, signum)

    def timeout_handler(self, timer_handle):
        """
        Callback that is triggered by the timer indicating a timeout.

        The timer is reset every time a condition is met.
        """
        self.close_handles()
        self.event_queue.put(
            message.TimeoutMessage(self.timeout, self.screenshot())
        )

    def on_exit(self, proc, exit_status, term_signal):
        """
        Callback that is triggered when the process exits of its
        own accord. Takes a screenshot and tells the control thread
        that we're done.
        """
        self.check_order_queue()
        self.close_handles()
        self.event_queue.put(
            message.ExitMessage(message.FinishedProcess(
                exit_status,
                term_signal,
                '\n'.join(self.screen.display),
            ))
        )

    def on_tty_read(self, handle, data, error):
        """
        Callback that is triggered when the process spews a chunk
        of data.
        """
        if data is None:
            pass
        else:
            self.stream.feed(data.decode('utf8'))
            self.raw_byte_output = self.raw_byte_output + data
            self.check_order_queue()

    def check_order_queue(self):
        """
        Callback that is triggered when there is (probably) a message
        waiting on the order queue.
        """
        try:
            if self.task is None:
                try:
                    self.task = self.order_queue.get(block=False)
                except queue.Empty:
                    self.task = None

            if self.task is not None:
                if isinstance(self.task, message.KillProcess):
                    for descendant in self.psutil.children(recursive=True):
                        descendant.kill()
                    self.psutil.kill()
                    self.close_handles()
                    self.event_queue.put(message.ProcessKilled())
                if isinstance(self.task, message.KeyPresses):
                    os.write(self.master_fd, self.task.value)
                    self.task = None
                elif isinstance(self.task, message.Condition):
                    self.reset_timeout(self.task.timeout)
                    iscreen = IScreen(self.screen, self.raw_byte_output)

                    if self.task.condition_function(iscreen):
                        self.event_queue.put(message.OutputMatched())
                        self.task = None
                elif isinstance(self.task, message.TakeScreenshot):
                    self.event_queue.put(message.Screenshot(
                        self.screenshot()
                    ))
                    self.task = None

        except Exception as error:
            self.close_handles()
            self.event_queue.put(message.ExceptionMessage(error))

    def reset_timeout(self, timeout=None):
        """
        Clears the timeout handle if it's there and sets a new one
        if timeout is set.
        """
        self.timeout = None
        if self.timeout_handle is not None:
            self.timeout_handle.close()
            self.timeout_handle = None
        if timeout is not None:
            self.timeout = timeout
            self.timeout_handle = pyuv.Timer(self.loop)
            self.timeout_handle.start(self.timeout_handler, timeout, 0)

    def close_handles(self):
        """
        Shut down all the callbacks.
        """
        if self.timeout_handle is not None and not self.timeout_handle.closed:
            self.timeout_handle.close()
            self.timeout_handle = None
        if not self.tty_handle.closed:
            self.tty_handle.close()
            self.tty_handle = None
        if not self.async_handle.closed:
            self.async_handle.close()
            self.async_handle = None
        if not self.process_handle.closed:
            self.process_handle.close()
            self.process_handle = None
        if not self.sigint_handle.closed:
            self.sigint_handle.close()
            self.sigint_handle = None
        if not self.sigterm_handle.closed:
            self.sigterm_handle.close()
            self.sigterm_handle = None





def stripshot(text):
    """
    Strips whitespace from right and bottom of
    a terminal screenshot.

    >>> stripshot(u'a\\nb\\n')
    'a\\nb'

    >>> stripshot(u'a   \\nb    \\n    \\n')
    'a\\nb'

    >>> stripshot(u'a   \\n     \\nc   \\n')
    'a\\n\\nc'
    """
    return u'\n'.join(
        line.rstrip() for line in text.split(u"\n") if line != u""
    ).rstrip()


"""
def pty_make_controlling_tty(tty_fd):

    This makes the pseudo-terminal the controlling tty. This should be
    more portable than the pty.fork() function. Specifically, this should
    work on Solaris.

    child_name = os.ttyname(tty_fd)

    # Disconnect from controlling tty. Harmless if not already connected.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        if fd >= 0:
            os.close(fd)
    except:
        # Already disconnected. This happens if running inside cron.
        pass

    # Verify we are disconnected from controlling tty
    # by attempting to open it again.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        if fd >= 0:
            os.close(fd)
            raise Exception('Failed to disconnect from ' +
                'controlling tty. It is still possible to open /dev/tty.')
    # which exception, shouldnt' we catch explicitly .. ?
    except:
        # Good! We are disconnected from a controlling tty.
        pass

    # Verify we can open child pty.
    fd = os.open(child_name, os.O_RDWR)
    if fd < 0:
        raise Exception("Could not open child pty, " + child_name)
    else:
        os.close(fd)

    # Verify we now have a controlling tty.
    fd = os.open("/dev/tty", os.O_WRONLY)
    if fd < 0:
        raise Exception("Could not open controlling tty, /dev/tty")
    else:
        os.close(fd)
"""

"""
#import array
#import fcntl
#import termios
#buf = array.array('h', [icommand.width, icommand.height, 0, 0])
#fcntl.ioctl(self.slave_fd, termios.TIOCSWINSZ, buf)
"""



"""
These are messages passed either to or from the main thread
to the handler thread along the response or request queue.

Master thread -> Order Queue -> Handler Thread
Handler thread -> Response Queue -> Master Thread
"""


class Message(object):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value


class PIDMessage(Message):
    pass


class FDMessage(Message):
    pass


class AsyncSendMethodMessage(Message):
    pass


class RunningProcess(object):
    def __init__(self, pid, stdin):
        self._pid = pid
        self._stdin = stdin


class FinishedProcess(object):
    def __init__(self, exit_code, term_signal, screenshot):
        self.exit_code = exit_code
        self.term_signal = term_signal
        self.screenshot = screenshot


class ExitMessage(Message):
    pass


class ExceptionMessage(Message):
    pass


class ProcessStartedMessage(Message):
    pass


class Condition(Message):
    def __init__(self, condition_function, timeout):
        self.condition_function = condition_function
        self.timeout = timeout


class OutputMatched(Message):
    def __init__(self):
        self._value = None


class TimeoutMessage(Message):
    def __init__(self, after, screenshot):
        self.after = after
        self.screenshot = screenshot


class KeyPresses(Message):
    pass


class TakeScreenshot(Message):
    def __init__(self):
        self._value = None


class Screenshot(Message):
    pass


class StartTimeout(Message):
    pass


class KillProcess(Message):
    def __init__(self):
        self._value = None


class ProcessKilled(Message):
    def __init__(self):
        self._value = None



from icommandlib.utils import stripshot


class ICommandError(Exception):
    pass


class IProcessTimeout(ICommandError):
    pass


class IProcessExitError(ICommandError):
    def __init__(self, exit_code, screenshot):
        self.exit_code = exit_code
        self.screenshot = screenshot

        super(IProcessExitError, self).__init__(
            self.MESSAGE.format(
                exit_code=self.exit_code,
                screenshot=stripshot(self.screenshot)
            )
        )


class UnexpectedExit(IProcessExitError):
    MESSAGE = (
        u"Process unexpectedly exited with "
        u"exit code {exit_code}. Output:\n{screenshot}"
    )


class AlreadyExited(IProcessExitError):
    MESSAGE = (
        u"Process already exited with {exit_code}. "
        u"Output:\n{screenshot}"
    )


class ExitWithError(IProcessExitError):
    MESSAGE = (
        u"Process exited with non-zero exit code {exit_code}. "
        u"Output:\n{screenshot}"
    )
